.\" A font suitable for Scheme keywords and program examples
.\" must be mounted on font position 5.
.\" For example:  .fp 5 TT  (Typewriter font)
.\"          or:  .fp 5 HR  (Helvetica Roman)
.\"
.fp 5 HR
.de f
.br
.ne 3
.ft 5
.nr x \w'\\$1'
.if \\n(.l<\\nx .nr x \\n(.l
.sp -.5
.ps 6
\l'\\nxu'
.ps
.sp -.1
\\$1
.ft
.LP
.if !'\\$2'' \f2[\\$2]\fP
.if !'\\$2'' .br
..
.de fa
.br
.ne 3
.ft 5
.nr x \w'\\$1'
.if \\n(.l<\\nx .nr x \\n(.l
.sp -.5
.ps 6
\l'\\nxu'
.ps
.sp -.1
\\$1
.ft
..
.de fb
.br
.ne 3
.ft 5
.sp -.1
\\$1
.ft
..
.de fc
.br
.ne 3
.ft 5
.sp -.1
\\$1
.ft
.LP
.if !'\\$2'' \f2[\\$2]\fP
.if !'\\$2' .br
..
.TL
Functions Exported by the Scheme/Xlib Interface
.AU
Oliver Laumann
.AB
This memo provides a list of the functions exported by the interface
to the Xlib library.
If a function is equivalent to a function of the C library, only
the name of the corresponding C function is given (in square brackets).
.AE
.NH
Loading the Interface
.LP
The Xlib-interface is loaded by evaluating
.DS
.ft 5
(require 'xlib).
.ft
.DE
.NH
Functions Operating on Colors

.f "(color? x)"
Returns #t if \f5x\fP is a color, #f otherwise.

.f "(make-color r g b)"
Returns an object of type \f5color\fP with the specified RGB components.
\f5r\fP, \f5g\fP, and \f5b\fP are reals in the range 0.0 to 1.0.

.f "(color-rgb-values color)"
Returns a list of three elements, the RGB components of the
given color (see \f5make-color\fP above).

.f "(query-color colormap pixel)" XQueryColor

.f "(query-colors colormap pixels)" XQueryColors
\f5pixels\fP is a vector of pixels.
Returns a vector of colors of the same size as \f5pixels\fP.

.f "(lookup-color colormap color-name)" XLookupColor
\f5color-name\fP is a string or a symbol.
Returns a pair of colors.

.NH
Functions Operating on Colormaps

.f "(colormap? x)"
Returns #t if \f5x\fP is a colormap, #f otherwise.

.f "(colormap-display colormap)"
Returns the display associated with the given colormap.

.f "(free-colormap colormap)" XFreeColormap

.NH
Functions Operating on Cursors

.f "(cursor? x)"
Returns #t if \f5x\fP is a cursor, #f otherwise.

.f "(cursor-display cursor)"
Returns the display associated with the given cursor.
   
.f "(free-cursor cursor)" XFreeCursor

.f "(create-cursor src-pixmap mask-pixmap x y foreground background)" XCreatePixmapCursor
\f5mask-pixmap\fP can be the symbol \f5none\fP.

.f "(create-glyph-cursor src-font source-char mask-font mask-char foreground background)" XCreateGlyphCursor
The display is obtained from \f5src-font\fP.
\f5src-char\fP and \f5mask-char\fP are integers.
\f5mask-font\fP can be the symbol \f5none\fP.

.f "(create-font-cursor display char)" XCreateGlyphCursor
Calls \f5create-glyph-cursor\fP with the font named ``cursor'', the
specified \f5source-char\fP, a \f5mask-char\fP of \f5(1+ char)\fP,
black foreground, and white background.

.f "(recolor-cursor cursor foreground background)" XRecolorCursor

.f "(define-cursor window cursor)"
Synonym for \f5(set-window-cursor! window cursor)\fP

.f "(undefine-cursor window)"
Synonym for \f5(set-window-cursor! window 'none)\fP

.NH
Functions Operating on Displays

.f "(display? x)"
Returns #t if \f5x\fP is a display, #f otherwise.

.f "(open-display . name-of-display)" XOpenDisplay
\f5name-of-display\fP is a string or a symbol.
If no name is specified, a NULL name will be passed to XOpenDisplay.

.f "(close-display display)" XCloseDisplay
Finalizes all objects associated with the display, then closes
the display.

.f "(display-root-window display)" DefaultRootWindow

.f "(display-colormap display)" DefaultColormap
Returns the default colormap of the display's default screen.

.f "(display-default-gcontext display)" DefaultGC
Returns the default GC of the display's default screen.

.fa "(display-width display)"
.fc "(display-height display)" "DisplayWidth, DisplayHeight"
Returns the width/height of the display's default screen.

.f "(display-flush-output display)" XFlush

.f "(display-wait-output display discard-events?)" XSync

.f "(set-input-focus display window revert-to time)" XSetInputFocus
\f5window\fP can be the symbol \f5pointer-root\fP.
\f5revert-to\fP is a symbol (\f5none\fP, \f5pointer-root\fP, or \f5parent\fP).
\f5time\fP is an integer or the symbol \f5now\fP.

.f "(input-focus display)" XGetInputFocus
Returns a pair the car of which is a window, and the cdr is a symbol
(\f5none\fP, \f5pointer-root\fP, or \f5parent\fP).

.f "(set-after-function! display procedure)" XSetAfterFunction
Returns the old after function.
If \f5procedure\fP is #f, the current after function is disassociated
from the display.

.f "(after-function display)"
Returns the after function currently associated with the given
display (#f if there is none).

.f "(synchronize display)"
Sets the display's after function to \f5display-wait-output\fP.

.NH
Event Handling

.f "(handle-events display-or-window . clauses)" "XNextEvent, XWindowEvent"
\f5handle-events\fP is a special form.
Each \f5clause\fP is of the form \f5(guard function)\fP; \f5guard\fP
is either an event name (a symbol, e.g. \f5key-press\fP or \f5exposure\fP),
a list of event names, or the symbol \f5else\fP.
\f5handle-events\fP gets the next event from the specified display
using XNextEvent or from the specified window using XWindowEvent.
Then the event type is matched against each event name in each guard
in order.
When a match occurs, the corresponding function is invoked with
the name of the event being dispatched (a symbol) and other, event
specific arguments (the \f2details\fP).
When no clause matches and an \f5else\fP clause is present, the function
from this clause is invoked.
\f5handle-events\fP loops until a function returns a value not
equal to #f in which case handle-events returns this value.

.NH
Functions Operating on Fonts

.f "(font? x)"
Returns #t if \f5x\fP is a font, #f otherwise.

.f "(font-display font)"
Returns the display associated with the given font.

.f "(open-font display font-name)" XLoadQueryFont
\f5font-name\fP is a string or a symbol.

.f "(close-font font)" XUnloadFont

.f "(font-name font)"
Returns the name of the specified font or #f if the name could
not be determined (e.g. when the font has been obtained by a call
to \f5gcontext-font\fP).

.f "(gcontext-font gcontext)"
Calls XQueryFont with the GC obtained by XGContextFromGC.
Only a limited number of functions can be applied to a font
returned by \f5gcontext-font\fP, since it has neither a name nor
a font-ID.

.f "(list-font-names display pattern)" XListFonts
\f5pattern\fP is a string or a symbol.
Returns a vector of font names (strings).

.f "(list-fonts display pattern)" XListFontsWithInfo
\f5pattern\fP is a string or a symbol.
Returns a vector of fonts.
These fonts are ``pseudo fonts'' which do not have a font-ID.
A pseudo font is loaded automatically and turned into a ``real''
font the first time it is passed to a function that makes use
of the font-ID.
   
.f "(font-info font)"
This function is not intended to be used by the application programmer.

.fa "(font-direction font)"
.fb "(font-min-byte2 font)"
.fb "(font-max-byte2 font)"
.fb "(font-min-byte1 font)"
.fb "(font-max-byte1 font)"
.fb "(font-all-chars-exist? font)"
.fb "(font-default-char font)"
.fb "(font-ascent font)"
.fc "(font-descent font)"
These functions return the font direction as a symbol (\f5left-to-right\fP
or \f5right-to-left\fP), the first and last character (as an integer),
the first and last row (integer), an indication whether all characters
have non-zero size (boolean), the default character (integer), and the
ascent and descent (integer) of the specified font.

.f "(char-info font index)"
This function is not intended to be used by the application programmer.
\f5index\fP is an integer or one of the symbols \f5min\fP and \f5max\fP.

.fa "(char-rbearing font index)"
.fb "(char-lbearing font index)"
.fb "(char-width font index)"
.fb "(char-ascent font index)"
.fc "(char-descent font index)"
These functions return the metrics of
the character specified by the integer \f5index\fP of the given font.
Each function returns an integer.
\f5font\fP can be a 1-byte as well as a 2-byte font.

.fa "(max-char-lbearing font)"
.fb "(max-char-rbearing font)"
.fb "(max-char-width font)"
.fb "(max-char-ascent font)"
.fc "(max-char-descent font)"
These functions return the maximum metrics over all characters
in the specified font.
Each function returns an integer.

.fa "(min-char-lbearing font)"
.fb "(min-char-rbearing font)"
.fb "(min-char-width font)"
.fb "(min-char-ascent font)"
.fc "(min-char-descent font)"
These functions return the minimum metrics over all characters
in the specified font.
Each function returns an integer.

.f "(font-properties font)"
Returns a vector of font properties; each element of the vector
is a pair consisting of the property name (an atom) and an
unsigned integer (the value of the property).

.f "(font-property font property-name)"
Returns the value of the specified property associated with the
specified font.
\f5property-name\fP is a string or a symbol.

.f "(font-path display)" XGetFontPath
Returns the current font path as a vector of strings.

.f "(set-font-path! display path)" XSetFontPath
\f5path\fP is a list; each element is a string or a symbol.

.NH
Functions Operating on Graphic Contexts

.f "(gcontext? x)"
Returns #t if \f5x\fP is a GC, #f otherwise.

.f "(gcontext-display gcontext)"
Returns the display associated with the given GC.

.f "(create-gcontext window tagged-vector)" XCreateGC
This function is not intended to be used by the application programmer.

.f "(make-gcontext . attributes)" XCreateGC
This macro is used to create a new GC.
Each \f5attribute\fP is a list of two elements, the first of which
is a symbol identifying the attribute, and the second element is
the value of the attribute.
The attributes can be specified in any order.
Attributes are \f5window\fP (mandatory) and all the attributes that can be
set by the \f5set-gcontext-\fP\f2attribute\fP\f5!\fP functions
below.

.f "(copy-gcontext gcontext window)" XCopyGC
Returns a copy of \f5gcontext\fP (associated with the specified window).

.f "(free-gcontext gcontext)" XFreeGC

.f "(query-best-size display width height shape)" XQueryBestSize
\f5shape\fP is a symbol (\f5cursor\fP, \f5tile\fP, or \f5stipple\fP).
Returns a pair of integers (result width and result height).

.fa "(query-best-cursor display width height)"
.fb "(query-best-tile display width height)"
.fc "(query-best-stipple display width height)" XQueryBestSize
Invokes \f5query-best-size\fP with the given arguments and a shape
of \f5cursor\fP, \f5tile\fP, or \f5stipple\fP, respectively.

.f "(change-gcontext gcontext tagged-vector)" XChangeGC
This function is not intended to be used by the application programmer.

.fa "(set-gcontext-function! gcontext value)"
.fb "(set-gcontext-plane-mask! gcontext value)"
.fb "(set-gcontext-foreground! gcontext value)"
.fb "(set-gcontext-background! gcontext value)"
.fb "(set-gcontext-line-width! gcontext value)"
.fb "(set-gcontext-line-style! gcontext value)"
.fb "(set-gcontext-cap-style! gcontext value)"
.fb "(set-gcontext-join-style! gcontext value)"
.fb "(set-gcontext-fill-style! gcontext value)"
.fb "(set-gcontext-fill-rule! gcontext value)"
.fb "(set-gcontext-arc-mode! gcontext value)"
.fb "(set-gcontext-tile! gcontext value)"
.fb "(set-gcontext-stipple! gcontext value)"
.fb "(set-gcontext-ts-x! gcontext value)"
.fb "(set-gcontext-ts-y! gcontext value)"
.fb "(set-gcontext-font! gcontext value)"
.fb "(set-gcontext-subwindow-mode! gcontext value)"
.fb "(set-gcontext-exposures! gcontext value)"
.fb "(set-gcontext-clip-x! gcontext value)"
.fb "(set-gcontext-clip-y! gcontext value)"
.fb "(set-gcontext-clip-mask! gcontext value)"
.fb "(set-gcontext-dash-offset! gcontext value)"
.fc "(set-gcontext-dashes! gcontext value)" "XChangeGC"
Sets the logical operation, plane mask, foreground and background pixel
value, line width and style, cap and join style, fill style and rule,
arc mode, tiling and stippling pixmap, tiling x- and y-origin, font,
subwindow mode, clipping x- and y-origin, clipping bitmap, and dashed line
information for the specified graphics context.
.LP
The \f5value\fP argument to \f5set-gcontext-function!\fP is a symbol
(\f5clear\fP, \f5and\fP, \f5and-reverse\fP, \f5copy\fP, \f5and-inverted\fP,
\f5no-op\fP, \f5xor\fP, \f5or\fP, \f5nor\fP, \f5equiv\fP, \f5invert\fP,
\f5or-reverse\fP, \f5copy-inverted\fP, \f5nand\fP, or \f5set\fP).
The argument to \f5set-gcontext-plane-mask!\fP, \f5set-gcontext-foreground!\fP,
and \f5set-gcontext-background!\fP is a pixel.
\f5set-gcontext-tile!\fP, \f5set-gcontext-stipple\fP, and
\f5set-gcontext-clip-mask!\fP expect a pixmap as the second argument.
The line style is a symbol (\f5solid\fP, \f5dash\fP, \f5double-dash\fP);
the cap style is a symbol (\f5not-last\fP, \f5butt\fP, \f5round\fP,
\f5projecting\fP); the join style is a symbol (\f5miter\fP, \f5round\fP,
\f5bevel\fP); the fill style is a symbol (\f5solid\fP, \f5tiled\fP,
\f5stippled\fP, \f5opaque-stippled\fP); the fill rule is a symbol
(\f5even-odd\fP, \f5winding\fP); the arc mode is a symbol (\f5chord\fP,
\f5pie-slice\fP).
The \f5value\fP argument to \f5set-gcontext-font!\fP must be a font.
The \f5value\fP argument to \f5set-gcontext-subwindow-mode!\fP is a
symbol (\f5clip-by-children\fP, \f5include-inferiors\fP).
\f5set-gcontext-exposures!\fP expects a boolean.
All other functions have an integer \f5value\fP argument.

.NH
Graphics Functions

.f "(clear-area window x y width height exposures?)" XClearArea

.f "(copy-area src-drawable gcontext src-x src-y width height dst-drawable dst-x dst-y)" XCopyArea

.f "(copy-plane src-drawable gcontext plane src-x src-y width height dst-drawable dst-x dst-y)" XCopyPlane
\f5plane\fP is an integer.
An error is signaled unless exactly one bit is set in \f5plane\fP.

.f "(draw-point drawable gcontext x y)" XDrawPoint

.f "(draw-points drawable gcontext vector-of-points relative?)" XDrawPoints
\f5vector-of-points\fP is a vector of pairs consisting of two integers
(the x- and y-coordinates).
If \f5relative?\fP is #t, CoordModePrevious
is passed to XDrawPoints, otherwise CoordModeOrigin is used.

.f "(draw-line drawable gcontext x1 y1 x2 y2)" XDrawLine

.f "(draw-lines drawable gcontext vector-of-points relative?)" XDrawLines
See \f5draw-points\fP above.

.f "(draw-segments drawable gcontext vector-of-points)" XDrawSegments
\f5vector-of-points\fP is a vector of lists of four integers
(x1, y1, x2, and y2).

.f "(draw-rectangle drawable gcontext x y width height)" XDrawRectangle

.f "(fill-rectangle drawable gcontext x y width height)" XFillRectangle

.f "(draw-rectangles drawable gcontext vector-of-rectangles)" XDrawRectangles
\f5vector-of-rectangles\fP is a vector of lists of four integers
(x, y, width, and height).

.f "(fill-rectangles drawable gcontext vector-of-rectangles)" XFillRectangles
See \f5draw-rectangles\fP above.

.f "(draw-arc drawable gcontext x y width height angle1 angle2)" XDrawArc

.f "(fill-arc drawable gcontext x y width height angle1 angle2)" XFillArc

.f "(draw-arcs drawable gcontext vector-of-data)" XDrawArcs
\f5vector-of-data\fP is a vector of lists of six integers
(x, y, width, height, angle1, and angle2).

.f "(fill-arcs drawable gcontext vector-of-data)" XFillArcs
See \f5draw-arcs\fP above.

.f "(fill-polygon drawable gcontext vector-of-points relative? shape)" XFillPolygon
See \f5draw-points\fP above.
\f5shape\fP is a symbol (\f5complex\fP, \f5non-convex\fP, or \f5convex\fP).

.NH
Keycodes and Keysyms

.fa "(display-min-keycode display)"
.fc "(display-max-keycode display)"
Returns the minimum/maximum keycode (an integer) for the given display.

.f "(display-keysyms-per-keycode display)"
Returns the number of keysyms per keycode for the given display.

.f "(string\(mi>keysym string)" XStringToKeysym
\f5string\fP is a string or a symbol.
Returns an integer if XStringToKeysym succeeds, #f otherwise.

.f "(keysym\(mi>string keysym)" XKeysymToString
\f5keysym\fP is an integer.
Returns #f if XKeysymToString fails.

.f "(keycode\(mi>keysym display keycode index)" XKeycodeToKeysym
\f5keycode\fP and \f5index\fP are integers.

.f "(keysym\(mi>keycode display keysym)" XKeysymToKeycode
\f5keysym\fP is an integer.

.f "(lookup-string display keycode mask)" XLookupString
\f5keycode\fP is an integer.
\f5mask\fP is a list of symbols (\f5shift\fP, \f5lock\fP, \f5control\fP,
\f5mod1\fP .. \f5mod5\fP, \f5button1\fP .. \f5button5\fP,
or \f5any-modifier\fP).

.NH
Functions Operating on Pixels

.f "(pixel? x)"
Returns #t if \f5x\fP is a pixel, #f otherwise.

.f "(pixel-value pixel)"
Returns the value of the pixel as an unsigned integer.

.fa "(black-pixel display)"
.fc "(white-pixel display)" "BlackPixel, WhitePixel"
Applies BlackPixel/WhitePixel to the display's default screen.

.NH
Functions Operating on Pixmaps

.f "(pixmap? x)"
Returns #t if \f5x\fP is a pixmap, #f otherwise.

.f "(pixmap-display pixmap)"
Returns the display associated with the pixmap.

.f "(free-pixmap pixmap)" XFreePixmap

.f "(create-pixmap drawable width height depth)" XCreatePixmap

.f "(create-bitmap-from-data window data width height)" XCreateBitmapFromData
\f5data\fP is a string.
\f5(* width height)\fP must not exceed the number of bits in \f5string\fP.

.f "(write-bitmap-file filename pixmap width height x-hot y-hot)" XWriteBitmapFile
\f5filename\fP is a string or a symbol.
\f5x-hot\fP and \f5y-hot\fP are optional
(\(mi1 is used if they are omitted), but either both or none of them must be given.
\f5write-bitmap-file\fP returns a symbol (\f5success\fP, \f5open-failed\fP,
\f5file-invalid\fP, or \f5no-memory\fP).

.NH
Pointer Handling, Grabs

.f "(grab-pointer window owner? events pointer-sync? keyboard-sync? confine-to-window cursor time)" XGrabPointer
\f5events\fP is a list of symbols (event mask names, such as \f5enter-window\fP,
\f5pointer-motion\fP, etc.).
\f5pointer-sync?\fP and \f5keyboard-sync?\fP determine whether synchronous
or asynchronous grab mode is to be used.
\f5time\fP is an integer or the symbol \f5now\fP (for CurrentTime).
\f5grab-pointer\fP returns a symbol (\f5success\fP, \f5not-viewable\fP,
\f5already-grabbed\fP, \f5frozen\fP, or \f5invalid-time\fP).

.f "(ungrab-pointer display time)" XUngrabPointer

.f "(grab-button window button modifiers owner? events pointer-sync? keyboard-sync? confine-to-window cursor)" XGrabButton
\f5button\fP is a symbol (\f5button1\fP .. \f5button5\fP, or \f5any-button\fP).
\f5modifiers\fP is a list of symbols (\f5shift\fP, \f5lock\fP, \f5control\fP,
\f5mod1\fP .. \f5mod5\fP, \f5button1\fP .. \f5button5\fP,
or \f5any-modifier\fP).
For the other arguments see \f5grab-pointer\fP above.

.f "(ungrab-button window button modifiers)" XUngrabButton
See \f5grab-button\fP above.

.f "(change-active-pointer-grab display events cursor time)" XChangeActivePointerGrab
\f5events\fP is a list of symbols (event mask names, such as \f5enter-window\fP,
\f5pointer-motion\fP, etc.).

.f "(grab-keyboard window owner? pointer-sync? keyboard-sync? time)" XGrabKeyboard
For a description of the arguments and the return value see
\f5grab-pointer\fP above.

.f "(ungrab-keyboard display time)" XUngrabKeyboard

.f "(grab-key window key modifiers owner? pointer-sync? keyboard-sync?)" XGrabKey
\f5key\fP is a keycode (an integer) or the symbol \f5any\fP.
For the other arguments see \f5grab-pointer\fP above.

.f "(ungrab-key window key modifiers)" XUngrabKey
See \f5grab-key\fP above.

.f "(allow-events display mode time)" XAllowEvents
\f5mode\fP is a symbol (\f5async-pointer\fP, \f5sync-pointer\fP,
\f5replay-pointer\fP, \f5async-keyboard\fP, \f5sync-keyboard\fP,
\f5replay-keyboard\fP, \f5async-both\fP, or \f5sync-both\fP).

.f "(grab-server display)" XGrabServer

.f "(ungrab-server display)" XUngrabServer

.f "(with-server-grabbed display . body-forms)"
This macro performs \f5grab-server\fP on the specified display,
evaluates the \f5body-forms\fP in order, and then ungrabs the server.
The macro body is guarded by a \f5dynamic-wind\fP to ensure that the
\f5ungrab-server\fP is performed when a body-form calls a continuation
created outside the macro, and that it is grabbed again when
the body is re-entered at a later point in time.
\f5with-server-grabbed\fP returns the value of the last body-form.

.f "(query-pointer window)" XQueryPointer
Returns a list of eight items:
  x
  y
  same-screen? (the return value of XQueryPointer, a boolean)
  root-window
  root-x
  root-y
  child-window
  modifiers (a list of modifier names, see \f5grab-button\fP above).

.f "(general-warp-pointer display dst-window dst-x dst-y src-window src-x src-y src-width src-height)" XWarpPointer

.f "(warp-pointer dst-window dst-x dst-y)" XWarpPointer
Invokes \f5general-warp-pointer\fP with the display associated with the
\f5dst-window\fP, the \f5dst-window\fP, \f5dst-x\fP, \f5dst-y\fP,
a \f5src-window\fP of \f5none\fP, and zero source coordinates and dimensions.

.f "(warp-pointer-relative display x-offset y-offset)" XWarpPointer
Invokes \f5general-warp-pointer\fP with the specified \f5display\fP,
a \f5dst-window\fP of \f5none\fP, \f5x-offset\fP, \f5y-offset\fP,
a \f5src-window\fP  of \f5none\fP, and zero source coordinates and dimensions.

.f "(bell display . percent)" XBell
If \f5percent\fP is omitted, 0 is used.

.NH
Functions Operating on Atoms and Properties

.f "(atom? x)"
Returns #t if \f5x\fP is an atom, #f otherwise.

.f "(make-atom value)"
Returns an atom with the given \f5value\fP (\f5value\fP is an integer).
This function is only used for debugging purposes.

.f "(intern-atom display name)" XInternAtom
\f5name\fP is a string or a symbol.
The atom is created if it does not yet exist.

.f "(find-atom display name)" XInternAtom
\f5name\fP is a string or a symbol.
If the atom does not exist, the symbol \f5none\fP is returned.

.f "(atom-name display atom)" XGetAtomName
Returns a string.

.f "(list-properties window)" XListProperties
Returns a vector of atoms.

.f "(get-property window property request-type offset length delete?)" XGetWindowProperty
\f5property\fP is an object of type \f5atom\fP.
\f5request-type\fP is an atom or #f in which case AnyPropertyType will be used.
\f5offset\fP and \f5length\fP are integers.
An error is signaled if XGetWindowProperty fails.
.LP
\f5get-property\fP returns a list of four items:
  the ``actual type'' (an atom)
  the format (an integer)
  the data (if any, () otherwise)
  the number of bytes left (an integer)
.LP
The data returned is either a string (if the format indicates
8-bit data) or a vector of integers.

.f "(change-property window property type format mode data)" XChangeProperty
\f5property\fP and \f5type\fP are atoms.
\f5format\fP is an integer (8, 16, or 32).
If \f5format\fP is 8 \f5data\fP must be a string, otherwise a vector of
integers of the appropriate size.
An error is signaled if the
value of \f5format\fP is invalid or if \f5data\fP holds an integer
that exceeds the size indicated by \f5format\fP.
\f5mode\fP is a symbol (\f5replace\fP, \f5prepend\fP, or \f5append\fP).

.f "(delete-property window property)" XDeleteProperty

.f "(rotate-properties window vector-of-atoms delta)" XRotateWindowProperties

.f "(set-selection-owner! display selection owner time)" XSetSelectionOwner
\f5selection\fP is an atom; \f5owner\fP is a window; \f5time\fP is an
integer or the symbol \f5now\fP (for CurrentTime).

.f "(selection-owner display selection)" XGetSelectionOwner

.f "(convert-selection selection target property requestor-window time)" XConvertSelection
\f5selection\fP and \f5target\fP are atoms;
\f5property\fP is an atom or the symbol \f5none\fP. 

.NH
Functions for Drawing Text

.f "(text-width font text format)" "XTextWidth, XTextWidth16"
\f5format\fP indicates whether 8-bit or 16-bit text is used; it is either
the symbol \f51-byte\fP or the symbol \f52-byte\fP.
\f5text\fP is a vector of integers; the integers must not exceed the
size indicated by the format.

.f "(text-extents font text format)" "XTextExtents, XTextExtents16"
This function is not intended to be used by the application programmer.

.fa "(extents-lbearing font text format)"
.fb "(extents-rbearing font text format)"
.fb "(extents-width font text format)"
.fb "(extents-ascent font text format)"
.fc "(extents-descent font text format) "XTextExtents, XTextExtents16""
These functions are used to compute the overall metrics of an 8-bit
or 16-bit character string.
Each function returns an integer.
For the format of \f5text\fP and \f5format\fP see \f5text-width\fP above.

.f "(draw-image-text drawable gcontext x y text format)" "XDrawImageString, XDrawImageString16"
See \f5text-width\fP above.

.f "(draw-poly-text drawable gcontext x y text format)" "XDrawText, XDrawText16"
See \f5text-width\fP above.
\f5text\fP is a vector of integers with intermixed objects of type \f5font\fP.

.f "(translate-text string)"
Convert the string into a representation suitable as an argument
to \f5text-width\fP, \f5draw-image-text\fP, or \f5draw-poly-text\fP
(a vector of integers obtained by applying \f5char\(mi>integer\fP
to the characters of the string argument).

.NH
Functions Operating on Windows

.f "(window? x)"
Returns #t if \f5x\fP is a window, #f otherwise.

.f "(drawable? x)"
Returns #t if \f5x\fP is a ``drawable'' (window or pixmap), #f otherwise.

.f "(window-display window)"
Returns the display associated with the window.

.f "(window-unique-id window)"
Returns a small integer uniquely identifying the given window.

.f "(create-window parent-window x y width height border-width tagged-vector)" XCreateWindow
This function is not intended to be used by the application programmer.

.f "(make-window . attributes)" XCreateWindow
This macro is used to create a new window.
Each \f5attribute\fP is a list of two elements, the first of which
is a symbol identifying the attribute, and the second element is
the value of the attribute.
The attributes can be specified in any order.
Attributes are \f5x\fP, \f5y\fP, \f5width\fP, \f5height\fP,
\f5border\fP (each of which has an integer value), \f5parent\fP
(the parent window), and all attributes that can be set by means
of the \f5set-window-\fP\f2attribute\fP\f5!\fP functions below
except \f5sibling\fP and \f5stack-mode\fP.
The attributes \f5parent\fP, \f5width\fP, and \f5height\fP are
mandatory.
The default for \f5x\fP and \f5y\fP is 0, the default for
\f5border\fP is 2.

.f "(configure-window window tagged-vector)" XConfigureWindow
This function is not intended to be used by the application programmer.

.f "(change-window-attributes window tagged-vector)" XChangeWindowAttributes
This function is not intended to be used by the application programmer.

.fa "(set-window-x! window value)"
.fb "(set-window-y! window value)"
.fb "(set-window-width! window value)"
.fb "(set-window-height! window value)"
.fb "(set-window-border-width! window value)"
.fb "(set-window-sibling! window value)"
.fb "(set-window-stack-mode! window value)"
.fb "(set-window-background-pixmap! window value)"
.fb "(set-window-background-pixel! window value)"
.fb "(set-window-border-pixmap! window value)"
.fb "(set-window-border-pixel! window value)"
.fb "(set-window-bit-gravity! window value)"
.fb "(set-window-gravity! window value)"
.fb "(set-window-backing-store! window value)"
.fb "(set-window-backing-planes! window value)"
.fb "(set-window-backing-pixel! window value)"
.fb "(set-window-save-under! window value)"
.fb "(set-window-event-mask! window value)"
.fb "(set-window-do-not-propagate-mask! window value)"
.fb "(set-window-override-redirect! window value)"
.fb "(set-window-colormap! window value)"
.fc "(set-window-cursor! window value)" "XConfigureWindow, XChangeWindowAttributes"
Set the sibling window, stacking mode, background pixmap, background
pixel, border pixel, cursor, and other attributes (see
the \f5window-\fP functions below) of the specified window.
.LP
The stacking mode is a symbol (\f5above\fP, \f5below\fP, \f5top-if\fP,
\f5bottom-if\fP, \f5opposite\fP).
The \f5value\fP argument to \f5set-window-sibling!\fP must be a window,
\f5set-window-background-pixmap!\fP expects a pixmap,
\f5set-window-background-pixel!\fP and \f5set-window-border-pixel!\fP
expect a pixel, and \f5set-window-cursor!\fP expects a cursor argument.
For the types of the \f5value\fP argument of the other functions
see the return values of the \f5window-\fP functions below.

.f "(get-window-attributes window)" XGetWindowAttributes
This function is not intended to be used by the application programmer.

.fa "(window-x window)"
.fb "(window-y window)"
.fb "(window-width window)"
.fb "(window-height window)"
.fb "(window-border-width window)"
.fb "(window-depth window)"
.fb "(window-visual window)"
.fb "(window-root window)"
.fb "(window-class window)"
.fb "(window-bit-gravity window)"
.fb "(window-gravity window)"
.fb "(window-backing-store window)"
.fb "(window-backing-planes window)"
.fb "(window-backing-pixel window)"
.fb "(window-save-under window)"
.fb "(window-colormap window)"
.fb "(window-map-installed window)"
.fb "(window-map-state window)"
.fb "(window-all-event-masks window)"
.fb "(window-your-event-mask window)"
.fb "(window-do-not-propagate-mask window)"
.fb "(window-override-redirect window)"
.fc "(window-screen window)" XGetWindowAttributes
Returns the x- and y-coordinates, width, height, border width,
depth, visual, root window, class, bit gravity, window gravity,
backing store availability, backing planes, backing pixel,
save under availability, colormap, colormap installation information,
map state, global event mask, local event mask, ``do-not-propagate'' mask,
override redirect attribute, and screen of the specified window.
.LP
\f5window-visual\fP and \f5window-screen\fP always return the empty
list in the current release of the software.
\f5window-root\fP returns a window.
\f5window-class\fP returns a symbol (\f5input-output\fP, \f5input-only\fP).
\f5window-bit-gravity\fP returns a symbol (\f5forget\fP, \f5north-west\fP, 
\f5north\fP, \f5north-east\fP, \f5west\fP, \f5center\fP, \f5east\fP, 
\f5south-west\fP, \f5south\fP, \f5south-east\fP, \f5static\fP).
\f5window-gravity\fP returns a symbol (same as \f5window-bit-gravity\fP
with \f5unmap\fP instead of \f5forget\fP).
\f5window-backing-store\fP returns a symbol (\f5not-useful\fP,
\f5when-mapped\fP, \f5always\fP).
\f5window-backing-planes\fP and \f5window-backing-pixel\fP return
a pixel.
\f5window-save-under\fP, \f5window-map-installed\fP and
\f5window-override-redirect\fP return #t or #f.
\f5window-colormap\fP returns a colormap.
\f5window-map-state\fP returns a symbol (\f5unmapped\fP,
\f5unviewable\fP, \f5viewable\fP).
\f5window-all-event-masks\fP, \f5window-your-event-mask\fP, and
\f5window-do-not-propagate-mask\fP return a list of symbols
(event mask names such as \f5enter-window\fP, \f5pointer-motion\fP, etc.).
All other functions return an integer.

.f "(get-geometry drawable)" XGetGeometry
This function is not intended to be used by the application programmer.

.fa "(drawable-root drawable)"
.fb "(drawable-x drawable)"
.fb "(drawable-y drawable)"
.fb "(drawable-width drawable)"
.fb "(drawable-height drawable)"
.fb "(drawable-border-width drawable)"
.fc "(drawable-depth drawable)" XGetGeometry
Returns the root window, x- and y-coordinates, width, height,
border width, and depth of the specified drawable.
\f5drawable-root\fP returns a window, all other functions return
an integer.

.f "(map-window window)" XMapWindow

.f "(unmap-window window)" XUnmapWindow

.f "(destroy-window window)" XDestroyWindow

.f "(destroy-subwindows window)" XDestroySubwindows

.f "(map-subwindows window)" XMapSubwindows

.f "(unmap-subwindows window)" XUnmapSubwindows

.f "(reparent-window window parent-window x y)" XReparentWindow

.f "(clear-window window)"
Performs a \f5clear-area\fP on the entire window.

.f "(query-tree window)" XQueryTree
Returns a list of three elements:
  root-window
  parent-window
  children (a vector of windows).
 
.f "(translate-coordinates src-window x y dst-window)" XTranslateCoordinates
Returns a list of three elements:
  dst-x
  dst-y
  child-window.

.NH
Window Manager Functions

.f "(wm-name window)" XFetchName
Returns a string or #f if XFetchName fails.

.f "(wm-icon-name window)" XGetIconName
Returns a string or #f if XGetIconName fails.

.f "(set-wm-name! window name)" XStoreName
\f5name\fP is a string or a symbol.

.f "(set-wm-icon-name! window name)" XSetIconName
\f5name\fP is a string or a symbol.

.f "(wm-class window)" XGetClassHint
Returns a pair (name and class) each component of which is either
a string or #f.

.f "(set-wm-class! window name class)" XSetClassHint
\f5name\fP and \f5class\fP are strings or symbols.

.f "(wm-command window)"
Returns the value of the WM_COMMAND property of the given window
as a list of strings.

.f "(set-wm-command! window command)" XSetCommand
\f5command\fP is a list; each element is either a string or a symbol.

.f "(wm-hints window)" XGetWMHints
This function is not intended to be used by the application programmer.

.f "(set-wm-hints! window tagged-vector)" XSetWMHints
This function is not intended to be used by the application programmer.

.f "(size-hints window property)" XGetSizeHints
This function is not intended to be used by the application programmer.

.f "(set-size-hints! window property tagged-vector)" XSetSizeHints
This function is not intended to be used by the application programmer.

.f "(icon-sizes window)" XGetIconSize
This function is not intended to be used by the application programmer.

.f "(set-icon-sizes! window tagged-vector)" XSetIconSize
This function is not intended to be used by the application programmer.

.f "(transient-for window)" XGetTransientForHint

.f "(set-transient-for! window property-window)" XSetTransientForHint

.NH
Miscellaneous Functions

.f "(with object . body-forms)"
\f5with\fP is a macro.
\f5object\fP must be a drawable or a font.
The \f5body-forms\fP are evaluated in order; \f5with\fP returns the value
of the last body-form.
.LP
Within the scope of the \f5with\fP, the first call to an accessor
function accessing \f5object\fP (such as \f5window-\fP\f2attribute\fP
or \f5font-\fP\f2attribute\fP) causes the result of the corresponding
Xlib function to be retained in a cache; subsequent calls just return
the value from the cache.
Likewise, calls to Xlib functions for mutator functions modifying
\f5object\fP (such as \f5set-window-\fP\f2attribute\fP\f5!\fP)
are delayed until exit of the \f5with\fP body or until an accessor
function is called and the cached data for this accessor function
has been invalidated by the call to a mutator function.
