.\" A font suitable for Scheme keywords and program examples
.\" must be mounted on font position 5.
.\" For example:  .fp 5 TT  (Typewriter font)
.\"          or:  .fp 5 HR  (Helvetica Roman)
.\"
.fp 5 HR
.de f
.br
.ne 3
.ft 5
.nr x \w'\\$1'
.sp -.5
.ps 6
\l'\\nxu'
.ps
.sp -.1
\\$1
.ft
.LP
.if !'\\$2'' \f2[\\$2]\fP
.if !'\\$2'' .br
..
.de fa
.br
.ne 3
.ft 5
.nr x \w'\\$1'
.sp -.5
.ps 6
\l'\\nxu'
.ps
.sp -.1
\\$1
.ft
..
.de fb
.br
.ne 3
.ft 5
.sp -.1
\\$1
.ft
.LP
.if !'\\$2'' \f2[\\$2]\fP
.if !'\\$2'' .br
..
.TL
Functions Exported by the Scheme/Xt Interface
.AU
Oliver Laumann
.AB
This memo provides a list of the functions exported by the interface
to the Xt toolkit.
If a function is equivalent to a function of the C library, only
the name of the corresponding C function is given (in square brackets).
.AE
.NH
Loading the Interface
.LP
The Xt-interface is loaded by evaluating
.DS C
.ft 5
(require 'xwidgets).
.ft
.DE
This loads the Xlib-interface as well.
Individual widgets are loaded by the macro
.DS C
.ft 5
(load-widgets . widget-class-names)
.ft
.DE
.LP
The arguments are symbols (not quoted since \f5load-widgets\fP is a macro).
The widgets are loaded from subdirectories of ``scm/lib''; the directories
are searched according to the value of the variable \f5widget-load-path\fP.
The default for \f5widget-load-path\fP is ``(xaw xhp)'',
i.e. when there are an
Athena widget and an HP widget of the same name, the Athena widget
will be loaded.
.NH
Callback Functions

.f "(add-callbacks widget callback-name . callback-functions)" XtAddCallbacks
Adds the functions to a callback list of the specified widget.
\f5callback-name\fP is a string or a symbol.

.f "(add-callback widget callback-name callback-function)"
Calls \f5add-callbacks\fP with the specified function.
.NH
Functions Operating on Widget Classes

.f "(class? x)"
Returns #t if \f5x\fP is a class, #f otherwise.

.f "(find-class name-of-class)"
Returns the widget class of the specified name.
The name is a string or a symbol.

.f "(class-resources widget-class)" XtGetResourceList
Returns the resource list of the specified widget class.
Each item in the list is a list of the resource name,
the resource class, and the resource type.

.f "(class-constraint-resources widget-class)" XtGetConstraintRespourceList
See \f5class-resources\fP above.

.f "(class-sub-resources widget-class)"
Returns the list of sub-resources (if there are any) for the
specified widget class.  See \f5class-resources\fP above.

.f "(class-exists? name-of-class)"
Returns #t if a widget class of the given name exists
(i.e. has been loaded by means of \f5load-widgets\fP), #f otherwise.
.NH
Application Contexts

.f "(context? x)"
Returns #t if \f5x\fP is a context, #f otherwise.

.f "(create-context)" XtCreateApplicationContext

.f "(destroy-context context)" XtDestroyApplicationContext

.f "(initialize-display context display application-name application-class)" "XtDisplayInitialize, XtOpenDisplay"
If \f5display\fP is an object of type \f5display\fP, XtDisplayInitialize is
called.
If \f5display\fP is a display name (string or symbol) or #f,
XtOpenDisplay is called (with a NULL display in the latter case),
and the newly opened display is returned.

.f "(context-main-loop context)" XtAppMainLoop

.f "(context-pending context)" XtAppPending
Returns a list of symbols (\f5x-event\fP, \f5timer\fP, \f5alternate-input\fP).

.f "(context-process-event context . mask)" XtProcessEvent
\f5mask\fP is a list of symbols (see \f5context-pending\fP above).
XtIMAll is used if no mask is given.

.f "(context-add-work-proc context procedure)" XtAppAddWorkProc
Returns an identifier that can be used as an argument to \f5remove-work-proc\fP.

.f "(remove-work-proc work-proc-identifier)" XtRemoveWorkProc

.f "(context-add-timeout context timer-value procedure)" XtAppAddTimeOut
\f5timer-value\fP is an integer.
Returns an identifier that can be used as an argument to \f5remove-timeout\fP.

.f "(remove-timeout timeout-identifier)" XtRemoveTimeOut

.f "(identifier? x)"
Returns #t if \f5x\fP is an identifier (an opaque object returned by functions
like \f5context-add-timeout\fP), #f otherwise.
.NH
Popup Shells

.fa "(create-popup-shell widget-class parent-widget . args)"
.fb "(create-popup-shell widget-name widget-class parent-widget . args)" XtCreatePopupShell
If no widget name is given, the name of the shell class is used.
The number of \f5args\fP must be even, the 1st, 3rd, etc. argument
is the name of a resource to be set (a string or a symbol); the
2nd, 4th, etc. argument is the corresponding value.

.f "(popup shell-widget grab-kind)" XtPopup
\f5grab-kind\fP is a symbol (\f5grab-once\fP, \f5grab-nonexclusive\fP,
or \f5grab-exclusive\fP).

.f "(popdown shell-widget)" XtPopdown
.NH
Translations Tables

.f "(augment-translations widget table)" XtAugmentTranslations
The translation table is a string; XtParseTranslationTable is
applied to it.

.f "(override-translations widget table)" XtOverrideTranslations
See \f5augment-translations\fP above.

.f "(uninstall-translations widget)" XtUninstallTranslations
.NH
Functions Operating on Widgets

.f "(widget? x)"
Returns #t if \f5x\fP is a widget, #f otherwise.

.f "(destroy-widget widget)" XtDestroyWidget

.f "(create-shell application-name application-class parent display . args)" XtAppCreateShell
\f5application-name\fP and \f5application-class\fP are strings or symbols.
For the format of \f5args\fP see \f5create-popup-shell\fP above.

.fa "(create-widget widget-class parent . args)"
.fb "(create-widget widget-name widget-class parent . args)" XtCreateWidget
If no \f5widget-name\fP is given, the name of the widget class is used.
For the format of \f5args\fP see \f5create-popup-shell\fP above.

.f "(create-managed-widget . args)"
Applies \f5create-widget\fP to the arguments and then calls \f5manage-child\fP
with the newly created widget.

.f "(realize-widget widget)" XtRealizeWidget

.f "(unrealize-widget widget)" XtUnrealizeWidget

.f "(widget-realized? widget)" XtIsRealized

.f "(widget-display widget)" XtDisplay

.f "(widget-parent widget)" XtParent

.f "(widget-window widget)" XtWindow

.f "(widget-composite? widget)" XtIsComposite

.f "(manage-children . widgets)" XtManageChildren

.f "(manage-child widget)"
Calls \f5manage-children\fP with the specified widget.

.f "(unmanage-children . widgets)" XtUnmanageChildren

.f "(unmanage-child widget)"
Calls \f5unmanage-children\fP with the specified widget.

.f "(widget-managed? widget)" XtIsManaged

.f "(widget-class widget)" XtClass

.f "(widget-superclass widget)" XtSuperclass
If the widget is an instance of ``widgetClass'', the symbol \f5none\fP
is returned.

.f "(widget-subclass? widget class)" XtIsSubclass

.f "(set-mapped-when-managed! widget managed?)" XtSetMappedWhenManaged

.f "(map-widget widget)" XtMapWidget

.f "(unmap-widget widget)" XtUnmapWidget

.f "(set-values! widget . args)" XtSetValues
For the format of \f5args\fP see \f5create-popup-shell\fP above.

.f "(get-values widget . args)" XtGetValues
Returns a list of the values of the specified widget arguments.

.f "(widget-context widget)" XtWidgetToApplicationContext

.f "(set-sensitive! widget sensitive?)" XtSetSensitive

.f "(widget-sensitive? widget)" XtIsSensitive

.f "(window\(mi>widget window)" XtWindowToWidget

.f "(name\(mi>widget root-widget name)" XtNameToWidget
\f5name\fP is a string or a symbol.
